from __future__ import annotations

from pykotor.gl.models.boundary import Boundary
from pykotor.gl.models.cube import Cube
from pykotor.gl.models.empty import Empty
from pykotor.gl.models.mesh import Mesh
from pykotor.gl.models.model import Model
from pykotor.gl.models.node import Node
from pykotor.gl.models.read_mdl import gl_load_mdl, gl_load_stitched_model

__all__ = [
    "Boundary",
    "Cube",
    "Empty",
    "Mesh",
    "Model",
    "Node",
    "gl_load_mdl",
    "gl_load_stitched_model",
]



"""GLM compatibility layer using NumPy for PyPy support.

This module provides a PyGLM-compatible API using NumPy arrays,
allowing PyKotorGL to work with PyPy where PyGLM is not available.

Note: All classes intentionally use lowercase names to match PyGLM's API.
Private _data members are intentionally accessed for internal operations.
"""
# ruff: noqa: SLF001

from __future__ import annotations

import math

from typing import TYPE_CHECKING, overload

import numpy as np

if TYPE_CHECKING:
    from numpy import ndarray

try:
    import glm
except ImportError:
    class vec3:  # noqa: N801
        """3D vector class compatible with PyGLM vec3."""

        def __init__(self, x: float = 0.0, y: float = 0.0, z: float = 0.0):
            if isinstance(x, (vec3, np.ndarray)):
                self._data: ndarray = np.array(x._data if isinstance(x, vec3) else x, dtype=np.float32)  # noqa: SLF001
            else:
                self._data = np.array([x, y, z], dtype=np.float32)

        @property
        def x(self) -> float:
            return float(self._data[0])

        @x.setter
        def x(self, value: float):
            self._data[0] = value

        @property
        def y(self) -> float:
            return float(self._data[1])

        @y.setter
        def y(self, value: float):
            self._data[1] = value

        @property
        def z(self) -> float:
            return float(self._data[2])

        @z.setter
        def z(self, value: float):
            self._data[2] = value

        def __repr__(self) -> str:
            return f"vec3({self.x}, {self.y}, {self.z})"

        def __eq__(self, other: object) -> bool:
            if not isinstance(other, vec3):
                return NotImplemented
            return bool(np.allclose(self._data, other._data))

        def __hash__(self) -> int:
            return hash(tuple(self._data.flatten()))


    class vec4:  # noqa: N801
        """4D vector class compatible with PyGLM vec4."""

        def __init__(self, x: float = 0.0, y: float = 0.0, z: float = 0.0, w: float = 0.0):
            if isinstance(x, (vec4, np.ndarray)):
                self._data: ndarray = np.array(x._data if isinstance(x, vec4) else x, dtype=np.float32)  # noqa: SLF001
            else:
                self._data = np.array([x, y, z, w], dtype=np.float32)

        @property
        def x(self) -> float:
            return float(self._data[0])

        @x.setter
        def x(self, value: float):
            self._data[0] = value

        @property
        def y(self) -> float:
            return float(self._data[1])

        @y.setter
        def y(self, value: float):
            self._data[1] = value

        @property
        def z(self) -> float:
            return float(self._data[2])

        @z.setter
        def z(self, value: float):
            self._data[2] = value

        @property
        def w(self) -> float:
            return float(self._data[3])

        @w.setter
        def w(self, value: float):
            self._data[3] = value

        def __repr__(self) -> str:
            return f"vec4({self.x}, {self.y}, {self.z}, {self.w})"

        def __eq__(self, other: object) -> bool:
            if not isinstance(other, vec4):
                return NotImplemented
            return bool(np.allclose(self._data, other._data))

        def __hash__(self) -> int:
            return hash(tuple(self._data.flatten()))


    class quat:  # noqa: N801
        """Quaternion class compatible with PyGLM quat."""

        def __init__(self, w: float = 1.0, x: float = 0.0, y: float = 0.0, z: float = 0.0):
            if isinstance(w, quat):
                self._data: ndarray = np.array(w._data, dtype=np.float32)  # noqa: SLF001
            elif isinstance(w, vec3):
                # vec3 - Euler angles
                self._data = self._from_euler(w)._data  # noqa: SLF001
            else:
                # w, x, y, z format
                self._data = np.array([w, x, y, z], dtype=np.float32)

        @staticmethod
        def _from_euler(euler: vec3) -> quat:
            """Create quaternion from Euler angles (in radians)."""
            cy = math.cos(euler.z * 0.5)
            sy = math.sin(euler.z * 0.5)
            cp = math.cos(euler.y * 0.5)
            sp = math.sin(euler.y * 0.5)
            cr = math.cos(euler.x * 0.5)
            sr = math.sin(euler.x * 0.5)

            w = cr * cp * cy + sr * sp * sy
            x = sr * cp * cy - cr * sp * sy
            y = cr * sp * cy + sr * cp * sy
            z = cr * cp * sy - sr * sp * cy

            return quat(w, x, y, z)

        @property
        def w(self) -> float:
            return float(self._data[0])

        @w.setter
        def w(self, value: float):
            self._data[0] = value

        @property
        def x(self) -> float:
            return float(self._data[1])

        @x.setter
        def x(self, value: float):
            self._data[1] = value

        @property
        def y(self) -> float:
            return float(self._data[2])

        @y.setter
        def y(self, value: float):
            self._data[2] = value

        @property
        def z(self) -> float:
            return float(self._data[3])

        @z.setter
        def z(self, value: float):
            self._data[3] = value

        def __repr__(self) -> str:
            return f"quat({self.w}, {self.x}, {self.y}, {self.z})"

        def __eq__(self, other: object) -> bool:
            if not isinstance(other, quat):
                return NotImplemented
            return bool(np.allclose(self._data, other._data))

        def __hash__(self) -> int:
            return hash(tuple(self._data.flatten()))


    class mat4:  # noqa: N801
        """4x4 matrix class compatible with PyGLM mat4."""

        def __init__(self, value: float | mat4 | np.ndarray = 1.0):
            if isinstance(value, (mat4, np.ndarray)):
                self._data: ndarray = np.array(value._data if isinstance(value, mat4) else value, dtype=np.float32)  # noqa: SLF001
            else:
                # Identity matrix scaled by value
                self._data = np.eye(4, dtype=np.float32) * value

        @overload
        def __mul__(self, other: mat4) -> mat4: ...

        @overload
        def __mul__(self, other: vec4) -> vec4: ...

        @overload
        def __mul__(self, other: vec3) -> vec3: ...

        def __mul__(self, other: mat4 | vec3 | vec4) -> mat4 | vec3 | vec4:
            if isinstance(other, mat4):
                result = mat4()
                result._data = np.matmul(self._data, other._data)
                return result
            if isinstance(other, vec4):
                result_data = np.matmul(self._data, other._data)
                return vec4(result_data[0], result_data[1], result_data[2], result_data[3])
            if isinstance(other, vec3):
                # Treat vec3 as vec4 with w=1
                vec4_data = np.array([other.x, other.y, other.z, 1.0], dtype=np.float32)
                result_data = np.matmul(self._data, vec4_data)
                return vec3(result_data[0], result_data[1], result_data[2])
            return NotImplemented

        def __repr__(self) -> str:
            return f"mat4(\n{self._data}\n)"

        def __eq__(self, other: object) -> bool:
            if not isinstance(other, mat4):
                return NotImplemented
            return bool(np.allclose(self._data, other._data))

        def __hash__(self) -> int:
            return hash(tuple(self._data.flatten()))


def translate(v: vec3) -> mat4:
    """Create a translation matrix."""
    result = mat4()
    result._data[3, 0] = v.x
    result._data[3, 1] = v.y
    result._data[3, 2] = v.z
    return result


def rotate(m: mat4, angle: float, axis: vec3) -> mat4:
    """Rotate a matrix by angle (radians) around axis."""
    # Normalize axis
    axis_length = math.sqrt(axis.x**2 + axis.y**2 + axis.z**2)
    if axis_length == 0:
        return mat4(m)

    x = axis.x / axis_length
    y = axis.y / axis_length
    z = axis.z / axis_length

    c = math.cos(angle)
    s = math.sin(angle)
    t = 1 - c

    # Rotation matrix
    rot = mat4()
    rot._data[0, 0] = t * x * x + c
    rot._data[0, 1] = t * x * y + s * z
    rot._data[0, 2] = t * x * z - s * y
    rot._data[1, 0] = t * x * y - s * z
    rot._data[1, 1] = t * y * y + c
    rot._data[1, 2] = t * y * z + s * x
    rot._data[2, 0] = t * x * z + s * y
    rot._data[2, 1] = t * y * z - s * x
    rot._data[2, 2] = t * z * z + c

    return m * rot


def mat4_cast(q: quat) -> mat4:
    """Convert a quaternion to a 4x4 rotation matrix."""
    result = mat4()

    qw, qx, qy, qz = q.w, q.x, q.y, q.z

    result._data[0, 0] = 1 - 2 * qy * qy - 2 * qz * qz
    result._data[0, 1] = 2 * qx * qy + 2 * qz * qw
    result._data[0, 2] = 2 * qx * qz - 2 * qy * qw

    result._data[1, 0] = 2 * qx * qy - 2 * qz * qw
    result._data[1, 1] = 1 - 2 * qx * qx - 2 * qz * qz
    result._data[1, 2] = 2 * qy * qz + 2 * qx * qw

    result._data[2, 0] = 2 * qx * qz + 2 * qy * qw
    result._data[2, 1] = 2 * qy * qz - 2 * qx * qw
    result._data[2, 2] = 1 - 2 * qx * qx - 2 * qy * qy

    return result


def inverse(m: mat4) -> mat4:
    """Calculate the inverse of a matrix."""
    result = mat4()
    try:
        result._data = np.linalg.inv(m._data)
    except np.linalg.LinAlgError:
        # Return identity matrix if singular
        result._data = np.eye(4, dtype=np.float32)
    return result


def perspective(fov: float, aspect: float, near: float, far: float) -> mat4:
    """Create a perspective projection matrix.

    Args:
    ----
        fov: Field of view in degrees
        aspect: Aspect ratio (width/height)
        near: Near clipping plane
        far: Far clipping plane

    Returns:
    -------
        mat4: Perspective projection matrix

    """
    result = mat4(0.0)

    fov_rad = math.radians(fov)
    tan_half_fov = math.tan(fov_rad / 2.0)

    result._data[0, 0] = 1.0 / (aspect * tan_half_fov)
    result._data[1, 1] = 1.0 / tan_half_fov
    result._data[2, 2] = -(far + near) / (far - near)
    result._data[2, 3] = -1.0
    result._data[3, 2] = -(2.0 * far * near) / (far - near)

    return result


def normalize(v: vec3) -> vec3:
    """Normalize a vector."""
    length = math.sqrt(v.x**2 + v.y**2 + v.z**2)
    if length == 0:
        return vec3(0, 0, 0)
    return vec3(v.x / length, v.y / length, v.z / length)


def cross(a: vec3, b: vec3) -> vec3:
    """Calculate the cross product of two vectors."""
    return vec3(
        a.y * b.z - a.z * b.y,
        a.z * b.x - a.x * b.z,
        a.x * b.y - a.y * b.x,
    )


def decompose(
    transform: mat4,
    scale: vec3,
    rotation: quat,
    translation: vec3,
    skew: vec3,  # noqa: ARG001
    perspective: vec4,  # noqa: ARG001
) -> bool:
    """Decompose a transformation matrix into its components.

    Args:
    ----
        transform: The transformation matrix to decompose
        scale: Output scale vector
        rotation: Output rotation quaternion
        translation: Output translation vector
        skew: Output skew vector (not implemented)
        perspective: Output perspective vector (not implemented)

    Returns:
    -------
        bool: True if decomposition was successful

    """
    m = transform._data

    # Extract translation
    translation.x = m[3, 0]
    translation.y = m[3, 1]
    translation.z = m[3, 2]

    # Extract scale
    scale_x = math.sqrt(m[0, 0]**2 + m[0, 1]**2 + m[0, 2]**2)
    scale_y = math.sqrt(m[1, 0]**2 + m[1, 1]**2 + m[1, 2]**2)
    scale_z = math.sqrt(m[2, 0]**2 + m[2, 1]**2 + m[2, 2]**2)

    scale.x = scale_x
    scale.y = scale_y
    scale.z = scale_z

    # Normalize matrix to extract rotation
    if scale_x == 0 or scale_y == 0 or scale_z == 0:
        rotation.w = 1.0
        rotation.x = 0.0
        rotation.y = 0.0
        rotation.z = 0.0
        return True

    rot_matrix = np.array([
        [m[0, 0] / scale_x, m[0, 1] / scale_x, m[0, 2] / scale_x],
        [m[1, 0] / scale_y, m[1, 1] / scale_y, m[1, 2] / scale_y],
        [m[2, 0] / scale_z, m[2, 1] / scale_z, m[2, 2] / scale_z],
    ], dtype=np.float32)

    # Convert rotation matrix to quaternion
    trace = rot_matrix[0, 0] + rot_matrix[1, 1] + rot_matrix[2, 2]

    if trace > 0:
        s = math.sqrt(trace + 1.0) * 2
        rotation.w = 0.25 * s
        rotation.x = (rot_matrix[2, 1] - rot_matrix[1, 2]) / s
        rotation.y = (rot_matrix[0, 2] - rot_matrix[2, 0]) / s
        rotation.z = (rot_matrix[1, 0] - rot_matrix[0, 1]) / s
    elif rot_matrix[0, 0] > rot_matrix[1, 1] and rot_matrix[0, 0] > rot_matrix[2, 2]:
        s = math.sqrt(1.0 + rot_matrix[0, 0] - rot_matrix[1, 1] - rot_matrix[2, 2]) * 2
        rotation.w = (rot_matrix[2, 1] - rot_matrix[1, 2]) / s
        rotation.x = 0.25 * s
        rotation.y = (rot_matrix[0, 1] + rot_matrix[1, 0]) / s
        rotation.z = (rot_matrix[0, 2] + rot_matrix[2, 0]) / s
    elif rot_matrix[1, 1] > rot_matrix[2, 2]:
        s = math.sqrt(1.0 + rot_matrix[1, 1] - rot_matrix[0, 0] - rot_matrix[2, 2]) * 2
        rotation.w = (rot_matrix[0, 2] - rot_matrix[2, 0]) / s
        rotation.x = (rot_matrix[0, 1] + rot_matrix[1, 0]) / s
        rotation.y = 0.25 * s
        rotation.z = (rot_matrix[1, 2] + rot_matrix[2, 1]) / s
    else:
        s = math.sqrt(1.0 + rot_matrix[2, 2] - rot_matrix[0, 0] - rot_matrix[1, 1]) * 2
        rotation.w = (rot_matrix[1, 0] - rot_matrix[0, 1]) / s
        rotation.x = (rot_matrix[0, 2] + rot_matrix[2, 0]) / s
        rotation.y = (rot_matrix[1, 2] + rot_matrix[2, 1]) / s
        rotation.z = 0.25 * s

    return True


def eulerAngles(q: quat) -> vec3:
    """Convert a quaternion to Euler angles (in radians).

    Args:
    ----
        q: Input quaternion

    Returns:
    -------
        vec3: Euler angles (roll, pitch, yaw) in radians

    """
    # Roll (x-axis rotation)
    sinr_cosp = 2 * (q.w * q.x + q.y * q.z)
    cosr_cosp = 1 - 2 * (q.x * q.x + q.y * q.y)
    roll = math.atan2(sinr_cosp, cosr_cosp)

    # Pitch (y-axis rotation)
    sinp = 2 * (q.w * q.y - q.z * q.x)
    if abs(sinp) >= 1:
        pitch = math.copysign(math.pi / 2, sinp)
    else:
        pitch = math.asin(sinp)

    # Yaw (z-axis rotation)
    siny_cosp = 2 * (q.w * q.z + q.x * q.y)
    cosy_cosp = 1 - 2 * (q.y * q.y + q.z * q.z)
    yaw = math.atan2(siny_cosp, cosy_cosp)

    return vec3(roll, pitch, yaw)


def value_ptr(obj: mat4 | vec3 | vec4) -> np.ndarray:
    """Get a pointer to the underlying data (returns flattened numpy array).

    Args:
    ----
        obj: Matrix or vector to get pointer from

    Returns:
    -------
        np.ndarray: Flattened array of the data

    """
    if isinstance(obj, mat4):
        # OpenGL expects column-major order
        return np.ascontiguousarray(obj._data.T.flatten())
    return np.ascontiguousarray(obj._data.flatten())


def unProject(
    win: vec3,
    model: mat4,
    proj: mat4,
    viewport: tuple[int, int, int, int],
) -> vec3:
    """Unproject a window coordinate to world coordinates.

    Args:
    ----
        win: Window coordinates (x, y, z where z is depth)
        model: Model-view matrix
        proj: Projection matrix
        viewport: Viewport as (x, y, width, height)

    Returns:
    -------
        vec3: World coordinates

    """
    # Compute the inverse of model * projection
    m: mat4 = proj * model  # type: ignore[assignment]
    inv_m: mat4 = inverse(m)

    # Normalize window coordinates to NDC [-1, 1]
    ndc = vec4(
        (win.x - viewport[0]) / viewport[2] * 2.0 - 1.0,
        (win.y - viewport[1]) / viewport[3] * 2.0 - 1.0,
        2.0 * win.z - 1.0,
        1.0,
    )

    # Transform NDC to world coordinates
    world: vec4 = inv_m * ndc  # type: ignore[assignment]

    # Perspective divide
    if world.w != 0:
        world.x /= world.w
        world.y /= world.w
        world.z /= world.w

    return vec3(world.x, world.y, world.z)




"""PyKotorGL - OpenGL rendering module for PyKotor.

Automatically uses PyGLM when available (CPython),
or falls back to NumPy-based glm_compat (PyPy).
"""

from __future__ import annotations

# Try to import PyGLM, fall back to our compatibility layer
try:
    import glm  # type: ignore[import-not-found]
    from glm import mat4, quat, vec3, vec4  # type: ignore[import-not-found]
    GLM_AVAILABLE = True
except ImportError:
    from pykotor.gl.glm_compat import (  # type: ignore[assignment]
        cross,
        decompose,
        eulerAngles,
        inverse,
        mat4,
        mat4_cast,
        normalize,
        perspective,
        quat,
        rotate,
        translate,
        unProject,
        value_ptr,
        vec3,
        vec4,
    )
    GLM_AVAILABLE = False

    # Create a glm module namespace for compatibility
    class _GLMNamespace:
        cross = staticmethod(cross)
        decompose = staticmethod(decompose)
        eulerAngles = staticmethod(eulerAngles)
        inverse = staticmethod(inverse)
        mat4_cast = staticmethod(mat4_cast)
        normalize = staticmethod(normalize)
        perspective = staticmethod(perspective)
        rotate = staticmethod(rotate)
        translate = staticmethod(translate)
        unProject = staticmethod(unProject)
        value_ptr = staticmethod(value_ptr)
        vec3 = vec3
        vec4 = vec4
        mat4 = mat4
        quat = quat

    glm = _GLMNamespace()  # type: ignore[assignment]

__all__ = [
    "GLM_AVAILABLE",
    "glm",
    "mat4",
    "quat",
    "vec3",
    "vec4",
]




from __future__ import annotations

import ctypes
import logging
import math
import struct

from copy import copy
from typing import TYPE_CHECKING

import numpy as np

from OpenGL import error as gl_error
from OpenGL.GL import glGenBuffers, glGenVertexArrays, glVertexAttribPointer
from OpenGL.GL.shaders import GL_FALSE
from OpenGL.raw.GL.ARB.tessellation_shader import GL_TRIANGLES
from OpenGL.raw.GL.ARB.vertex_shader import GL_FLOAT
from OpenGL.raw.GL.VERSION.GL_1_0 import GL_UNSIGNED_SHORT
from OpenGL.raw.GL.VERSION.GL_1_1 import glDrawElements
from OpenGL.raw.GL.VERSION.GL_1_3 import GL_TEXTURE0, GL_TEXTURE1, glActiveTexture
from OpenGL.raw.GL.VERSION.GL_1_5 import GL_ARRAY_BUFFER, GL_ELEMENT_ARRAY_BUFFER, GL_STATIC_DRAW, glBindBuffer, glBufferData
from OpenGL.raw.GL.VERSION.GL_2_0 import glEnableVertexAttribArray
from OpenGL.raw.GL.VERSION.GL_3_0 import glBindVertexArray

from pykotor.gl import glm, mat4, quat, vec3, vec4
from utility.common.geometry import Vector3

if TYPE_CHECKING:
    from pykotor.gl.scene import Scene
    from pykotor.gl.shader import Shader


logger = logging.getLogger(__name__)


class Model:
    def __init__(self, scene: Scene, root: Node):
        self._scene: Scene = scene
        self.root: Node = root

    def draw(
        self,
        shader: Shader,
        transform: mat4,
        *,
        override_texture: str | None = None,
    ):
        self.root.draw(shader, transform, override_texture)

    def find(self, name: str) -> Node | None:
        nodes: list[Node] = [self.root]
        while nodes:
            node: Node = nodes.pop()
            if node.name.lower() == name.lower():
                return node
            nodes.extend(node.children)
        return None

    def all(self) -> list[Node]:
        all_nodes: list[Node] = []
        search: list[Node] = [self.root]
        while search:
            node: Node = search.pop()
            search.extend(node.children)
            all_nodes.append(node)
        return all_nodes

    def box(self) -> tuple[vec3, vec3]:
        return self.bounds(mat4())

    def bounds(self, transform: mat4) -> tuple[vec3, vec3]:
        """Calculate the bounding box of the model with the given transform."""
        min_point = vec3(100000, 100000, 100000)
        max_point = vec3(-100000, -100000, -100000)
        self._box_rec(self.root, transform, min_point, max_point)

        min_point.x -= 0.1
        min_point.y -= 0.1
        min_point.z -= 0.1
        max_point.x += 0.1
        max_point.y += 0.1
        max_point.z += 0.1

        return min_point, max_point

    def _box_rec(
        self,
        node: Node,
        transform: mat4,
        min_point: vec3,
        max_point: vec3,
    ):
        """Calculates bounding box of node and its children recursively.

        Call the 'box' function to get started here, don't call this directly.

        Args:
        ----
            node: {Node object whose bounding box is calculated}
            transform: {Transformation matrix to apply on node}
            min_point: {vec3 to store minimum point of bounding box}
            max_point: {vec3 to store maximum point of bounding box}.

        Processing Logic:
        ----------------
            - Apply transformation on node position and rotation
            - Iterate through vertices of node mesh if present
            - Transform vertices and update bounding box points
            - Recursively call function for each child node.
        """
        transform = transform * glm.translate(node._position)  # noqa: SLF001
        transform = transform * glm.mat4_cast(node._rotation)  # noqa: SLF001

        if node.mesh and node.render:
            vertex_count = len(node.mesh.vertex_data) // node.mesh.mdx_size
            for i in range(vertex_count):
                index = i * node.mesh.mdx_size + node.mesh.mdx_vertex
                data = node.mesh.vertex_data[index : index + 12]
                x, y, z = struct.unpack("fff", data)
                position = transform * vec3(x, y, z)
                min_point.x = min(min_point.x, position.x)
                min_point.y = min(min_point.y, position.y)
                min_point.z = min(min_point.z, position.z)
                max_point.x = max(max_point.x, position.x)
                max_point.y = max(max_point.y, position.y)
                max_point.z = max(max_point.z, position.z)

        for child in node.children:
            self._box_rec(child, transform, min_point, max_point)


class Node:
    def __init__(
        self,
        scene: Scene,
        parent: Node | None,
        name: str,
    ):
        self._scene: Scene = scene
        self._parent: Node | None = parent
        self.name: str = name
        self._transform: mat4 = mat4()
        self._position: vec3 = glm.vec3()
        self._rotation: quat = glm.quat()
        self.children: list[Node] = []
        self.render: bool = True
        self.mesh: Mesh | None = None

        self._recalc_transform()

    def root(self) -> Node | None:
        ancestor: Node | None = self._parent
        while ancestor:
            ancestor = ancestor._parent  # noqa: SLF001
        return ancestor

    def ancestors(self) -> list[Node]:
        ancestors: list[Node] = []
        ancestor: Node | None = self._parent
        while ancestor:
            ancestors.append(ancestor)
            ancestor = ancestor._parent  # noqa: SLF001
        return list(reversed(ancestors))

    def global_position(self) -> vec3:  # sourcery skip: class-extract-method
        ancestors: list[Node] = [*self.ancestors(), self]
        transform = mat4()
        for ancestor in ancestors:
            transform = transform * glm.translate(ancestor._position)  # noqa: SLF001
            transform = transform * glm.mat4_cast(ancestor._rotation)  # noqa: SLF001
        position = vec3()
        glm.decompose(transform, vec3(), quat(), position, vec3(), vec4())  # pyright: ignore[reportCallIssue, reportArgumentType]
        return position

    def global_rotation(self) -> quat:
        ancestors: list[Node] = [*self.ancestors(), self]
        transform = mat4()
        for ancestor in ancestors:
            transform = transform * glm.translate(ancestor._position)  # noqa: SLF001
            transform = transform * glm.mat4_cast(ancestor._rotation)  # noqa: SLF001
        rotation = quat()
        glm.decompose(transform, vec3(), rotation, vec3(), vec3(), vec4())  # pyright: ignore[reportCallIssue, reportArgumentType]
        return rotation

    def global_transform(self) -> mat4:
        ancestors: list[Node] = [*self.ancestors(), self]
        transform = mat4()
        for ancestor in ancestors:
            transform = transform * glm.translate(ancestor._position)  # noqa: SLF001
            transform = transform * glm.mat4_cast(ancestor._rotation)  # noqa: SLF001
        return transform

    def transform(self) -> mat4:
        return copy(self._transform)

    def _recalc_transform(self):
        self._transform = glm.translate(self._position) * glm.mat4_cast(quat(self._rotation))

    def position(self) -> vec3:
        return copy(self._position)

    def set_position(self, x: float, y: float, z: float):
        self._position = vec3(x, y, z)
        self._recalc_transform()

    def rotation(self) -> quat:
        return copy(self._rotation)

    def set_rotation(
        self,
        pitch: float,
        yaw: float,
        roll: float,
    ):
        self._rotation = quat(vec3(pitch, yaw, roll))
        self._recalc_transform()

    def draw(
        self,
        shader: Shader,
        transform: mat4,
        override_texture: str | None = None,
    ):
        transform = transform * self._transform

        if self.mesh and self.render:
            self.mesh.draw(shader, transform, override_texture)

        for child in self.children:
            child.draw(shader, transform, override_texture=override_texture)


class Mesh:
    def __init__(
        self,
        scene: Scene,
        node: Node,
        texture: str,
        lightmap: str,
        vertex_data: bytearray,
        element_data: bytearray,
        block_size: int,
        data_bitflags: int,
        vertex_offset: int,
        normal_offset: int,
        texture_offset: int,
        lightmap_offset: int,
    ):
        self._scene: Scene = scene
        self._node: Node = node

        self.texture: str = "NULL"
        self.lightmap: str = "NULL"

        self.vertex_data: bytearray = vertex_data
        self.mdx_size: int = block_size
        self.mdx_vertex: int = vertex_offset
        self.mdx_texture: int = texture_offset
        self.mdx_lightmap: int = lightmap_offset
        self._index_data: bytes = bytes(element_data)
        self._vertex_blob_cache: bytes | None = None

        self._vao: int = glGenVertexArrays(1)
        self._vbo: int = glGenBuffers(1)
        self._ebo: int = glGenBuffers(1)
        glBindVertexArray(self._vao)

        glBindBuffer(GL_ARRAY_BUFFER, self._vbo)
        # Convert vertex_data bytearray to MemoryView
        vertex_data_mv = memoryview(vertex_data)
        glBufferData(GL_ARRAY_BUFFER, len(vertex_data), vertex_data_mv, GL_STATIC_DRAW)

        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, self._ebo)
        # Convert element_data bytearray to MemoryView
        element_data_mv = memoryview(element_data)
        glBufferData(GL_ELEMENT_ARRAY_BUFFER, len(element_data), element_data_mv, GL_STATIC_DRAW)

        self._face_count: int = len(element_data) // 2

        if data_bitflags & 0x0001:
            glEnableVertexAttribArray(1)
            glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, block_size, ctypes.c_void_p(vertex_offset))

        if data_bitflags & 0x0020 and texture and texture != "NULL":
            glEnableVertexAttribArray(3)
            glVertexAttribPointer(3, 2, GL_FLOAT, GL_FALSE, block_size, ctypes.c_void_p(texture_offset))
            self.texture = texture

        if data_bitflags & 0x0004 and lightmap and lightmap != "NULL":
            glEnableVertexAttribArray(4)
            glVertexAttribPointer(4, 2, GL_FLOAT, GL_FALSE, block_size, ctypes.c_void_p(lightmap_offset))
            self.lightmap = lightmap

        glBindBuffer(GL_ARRAY_BUFFER, 0)
        glBindVertexArray(0)

    def draw(
        self,
        shader: Shader,
        transform: mat4,
        override_texture: str | None = None,
    ):
        shader.set_matrix4("model", transform)

        glActiveTexture(GL_TEXTURE0)
        self._scene.texture(override_texture or self.texture).use()

        glActiveTexture(GL_TEXTURE1)
        self._scene.texture(self.lightmap, lightmap=True).use()

        glBindVertexArray(self._vao)
        glDrawElements(GL_TRIANGLES, self._face_count, GL_UNSIGNED_SHORT, None)

    def vertex_blob(self) -> bytes:
        """Generate vertex blob for ModernGL rendering.
        
        Returns a bytes object containing interleaved vertex data:
        - 3 floats for position (x, y, z)
        - 2 floats for diffuse UV (u, v)
        - 2 floats for lightmap UV (u, v)
        """
        if self._vertex_blob_cache is not None:
            return self._vertex_blob_cache

        import numpy as np

        vertex_count = len(self.vertex_data) // self.mdx_size
        if vertex_count == 0:
            self._vertex_blob_cache = b""
            return self._vertex_blob_cache

        blob = np.zeros((vertex_count, 7), dtype=np.float32)
        positions = np.frombuffer(
            self.vertex_data,
            dtype="<f4",
            count=vertex_count * 3,
            offset=self.mdx_vertex,
        ).reshape(vertex_count, 3)
        blob[:, 0:3] = positions

        if self.mdx_texture >= 0:
            diffuse = np.frombuffer(
                self.vertex_data,
                dtype="<f4",
                count=vertex_count * 2,
                offset=self.mdx_texture,
            ).reshape(vertex_count, 2)
            blob[:, 3:5] = diffuse

        if self.mdx_lightmap >= 0:
            lightmap = np.frombuffer(
                self.vertex_data,
                dtype="<f4",
                count=vertex_count * 2,
                offset=self.mdx_lightmap,
            ).reshape(vertex_count, 2)
            blob[:, 5:7] = lightmap

        self._vertex_blob_cache = blob.tobytes()
        return self._vertex_blob_cache

    @property
    def index_data(self) -> bytes:
        """Return the index data for the mesh."""
        return self._index_data


class Cube:
    def __init__(
        self,
        scene: Scene,
        min_point: vec3 | None = None,
        max_point: vec3 | None = None,
    ):
        self._scene = scene

        min_point = vec3(-1.0, -1.0, -1.0) if min_point is None else min_point
        max_point = vec3(1.0, 1.0, 1.0) if max_point is None else max_point

        vertices = np.array(
            [
                min_point.x,
                min_point.y,
                max_point.z,
                max_point.x,
                min_point.y,
                max_point.z,
                max_point.x,
                max_point.y,
                max_point.z,
                min_point.x,
                max_point.y,
                max_point.z,
                min_point.x,
                min_point.y,
                min_point.z,
                max_point.x,
                min_point.y,
                min_point.z,
                max_point.x,
                max_point.y,
                min_point.z,
                min_point.x,
                max_point.y,
                min_point.z,
            ],
            dtype="float32",
        )

        elements = np.array(
            [0, 1, 2, 2, 3, 0, 1, 5, 6, 6, 2, 1, 7, 6, 5, 5, 4, 7, 4, 0, 3, 3, 7, 4, 4, 5, 1, 1, 0, 4, 3, 2, 6, 6, 7, 3],
            dtype="int16",
        )

        self.min_point: vec3 = min_point
        self.max_point: vec3 = max_point

        self._vao = 0
        self._vbo = 0
        self._ebo = 0
        self._face_count: int = len(elements)
        self._buffers_supported = False

        try:
            self._vao = glGenVertexArrays(1)
            self._vbo = glGenBuffers(1)
            self._ebo = glGenBuffers(1)
            glBindVertexArray(self._vao)

            glBindBuffer(GL_ARRAY_BUFFER, self._vbo)
            glBufferData(GL_ARRAY_BUFFER, len(vertices) * 4, vertices, GL_STATIC_DRAW)

            glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, self._ebo)
            glBufferData(GL_ELEMENT_ARRAY_BUFFER, len(elements) * 4, elements, GL_STATIC_DRAW)

            glEnableVertexAttribArray(1)
            glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, 12, ctypes.c_void_p(0))

            glBindBuffer(GL_ARRAY_BUFFER, 0)
            glBindVertexArray(0)
            self._buffers_supported = True
        except gl_error.NullFunctionError:
            logger.warning(
                "OpenGL buffer objects are unavailable; falling back to CPU-only cube bounds."
            )
            self._face_count = 0
            self._vao = 0
            self._vbo = 0
            self._ebo = 0


    def draw(self, shader: Shader, transform: mat4):
        if not self._buffers_supported:
            return

        shader.set_matrix4("model", transform)
        glBindVertexArray(self._vao)
        glDrawElements(GL_TRIANGLES, self._face_count, GL_UNSIGNED_SHORT, None)


class Boundary:
    def __init__(
        self,
        scene: Scene,
        vertices: list[Vector3],
    ):
        self._scene: Scene = scene

        vertices, elements = self._build_nd(vertices)

        self._vao = 0
        self._vbo = 0
        self._ebo = 0
        self._face_count: int = len(elements)
        self._buffers_supported = False

        try:
            self._vao = glGenVertexArrays(1)
            self._vbo = glGenBuffers(1)
            self._ebo = glGenBuffers(1)
            glBindVertexArray(self._vao)

            glBindBuffer(GL_ARRAY_BUFFER, self._vbo)
            glBufferData(GL_ARRAY_BUFFER, len(vertices) * 4, vertices, GL_STATIC_DRAW)

            glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, self._ebo)
            glBufferData(GL_ELEMENT_ARRAY_BUFFER, len(elements) * 4, elements, GL_STATIC_DRAW)

            glEnableVertexAttribArray(1)
            glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, 12, ctypes.c_void_p(0))

            glBindBuffer(GL_ARRAY_BUFFER, 0)
            glBindVertexArray(0)
            self._buffers_supported = True
        except gl_error.NullFunctionError:
            logger.warning(
                "OpenGL buffer objects are unavailable; boundary rendering disabled."
            )
            self._face_count = 0
            self._vao = 0
            self._vbo = 0
            self._ebo = 0

    @classmethod
    def from_circle(
        cls,
        scene: Scene,
        radius: float,
        smoothness: int = 10,
    ) -> Boundary:
        vertices: list[Vector3] = []
        for i in range(smoothness):
            x = math.cos(i / smoothness * math.pi / 2)
            y = math.sin(i / smoothness * math.pi / 2)
            vertices.append(Vector3(x, y, 0) * radius)
        for i in range(smoothness):
            x = math.cos(i / smoothness * math.pi / 2 + math.pi / 2)
            y = math.sin(i / smoothness * math.pi / 2 + math.pi / 2)
            vertices.append(Vector3(x, y, 0) * radius)
        for i in range(smoothness):
            x = math.cos(i / smoothness * math.pi / 2 + math.pi / 2 * 2)
            y = math.sin(i / smoothness * math.pi / 2 + math.pi / 2 * 2)
            vertices.append(Vector3(x, y, 0) * radius)
        for i in range(smoothness):
            x = math.cos(i / smoothness * math.pi / 2 + math.pi / 2 * 3)
            y = math.sin(i / smoothness * math.pi / 2 + math.pi / 2 * 3)
            vertices.append(Vector3(x, y, 0) * radius)
        return Boundary(scene, vertices)

    def draw(self, shader: Shader, transform: mat4):
        if not self._buffers_supported:
            return

        shader.set_matrix4("model", transform)
        glBindVertexArray(self._vao)
        glDrawElements(GL_TRIANGLES, self._face_count, GL_UNSIGNED_SHORT, None)

    def _build_nd(self, vertices: list[Vector3]) -> tuple[np.ndarray, np.ndarray]:
        npvertices = []
        for vertex in vertices:
            npvertices.extend([*vertex, *Vector3(vertex.x, vertex.y, vertex.z + 2)])

        npfaces: list[int] = []
        count = len(vertices) * 2
        for i, _vertex in enumerate(vertices):
            index1 = i * 2
            index2 = i * 2 + 2 if i * 2 + 2 < count else 0
            index3 = i * 2 + 1
            index4 = (i * 2 + 2) + 1 if (i * 2 + 2) + 1 < count else 1
            npfaces.extend([index1, index2, index3, index2, index4, index3])
        return np.array(npvertices, dtype="float32"), np.array(npfaces, dtype="int16")


class Empty:
    def __init__(self, scene: Scene):
        self._scene: Scene = scene

    def draw(self, shader: Shader, transform: mat4):
        ...
